/*
   igraph library.
   Copyright (C) 2009-2025  The igraph development team <igraph@igraph.org>

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

#ifndef IGRAPH_COMMUNITY_H
#define IGRAPH_COMMUNITY_H

#include "igraph_decls.h"

#include "igraph_arpack.h"
#include "igraph_constants.h"
#include "igraph_datatype.h"
#include "igraph_error.h"
#include "igraph_types.h"
#include "igraph_vector_list.h"

IGRAPH_BEGIN_C_DECLS

/* -------------------------------------------------- */
/* K-Cores and K-Truss                                */
/* -------------------------------------------------- */

IGRAPH_EXPORT igraph_error_t igraph_coreness(
    const igraph_t *graph, igraph_vector_int_t *cores, igraph_neimode_t mode);
IGRAPH_EXPORT igraph_error_t igraph_trussness(
    const igraph_t* graph, igraph_vector_int_t* trussness);

/* -------------------------------------------------- */
/* Community Structure                                */
/* -------------------------------------------------- */

/* TODO: cut.community */
/* TODO: edge.type.matrix */
/* TODO:  */

IGRAPH_EXPORT igraph_error_t igraph_community_optimal_modularity(const igraph_t *graph,
                                                                 const igraph_vector_t *weights,
                                                                 igraph_real_t resolution,
                                                                 igraph_real_t *modularity,
                                                                 igraph_vector_int_t *membership);

IGRAPH_EXPORT igraph_error_t igraph_community_spinglass(const igraph_t *graph,
                                             const igraph_vector_t *weights,
                                             igraph_real_t *modularity,
                                             igraph_real_t *temperature,
                                             igraph_vector_int_t *membership,
                                             igraph_vector_int_t *csize,
                                             igraph_int_t spins,
                                             igraph_bool_t parupdate,
                                             igraph_real_t starttemp,
                                             igraph_real_t stoptemp,
                                             igraph_real_t coolfact,
                                             igraph_spincomm_update_t update_rule,
                                             igraph_real_t gamma,
                                             igraph_spinglass_implementation_t implementation,
                                             igraph_real_t gamma_minus);

IGRAPH_EXPORT igraph_error_t igraph_community_spinglass_single(const igraph_t *graph,
                                                    const igraph_vector_t *weights,
                                                    igraph_int_t vertex,
                                                    igraph_vector_int_t *community,
                                                    igraph_real_t *cohesion,
                                                    igraph_real_t *adhesion,
                                                    igraph_real_t *inner_links,
                                                    igraph_real_t *outer_links,
                                                    igraph_int_t spins,
                                                    igraph_spincomm_update_t update_rule,
                                                    igraph_real_t gamma);

IGRAPH_EXPORT igraph_error_t igraph_community_walktrap(const igraph_t *graph,
                                            const igraph_vector_t *weights,
                                            igraph_int_t steps,
                                            igraph_matrix_int_t *merges,
                                            igraph_vector_t *modularity,
                                            igraph_vector_int_t *membership);

IGRAPH_EXPORT igraph_error_t igraph_community_infomap(const igraph_t *graph,
                                           const igraph_vector_t *edge_weights,
                                           const igraph_vector_t *vertex_weights,
                                           igraph_int_t nb_trials,
                                           igraph_bool_t is_regularized,
                                           igraph_real_t regularization_strength,
                                           igraph_vector_int_t *membership,
                                           igraph_real_t *codelength);

IGRAPH_EXPORT igraph_error_t igraph_community_edge_betweenness(const igraph_t *graph,
                                                    igraph_vector_int_t *removed_edges,
                                                    igraph_vector_t *edge_betweenness,
                                                    igraph_matrix_int_t *merges,
                                                    igraph_vector_int_t *bridges,
                                                    igraph_vector_t *modularity,
                                                    igraph_vector_int_t *membership,
                                                    igraph_bool_t directed,
                                                    const igraph_vector_t *weights,
                                                    const igraph_vector_t *lengths);
IGRAPH_EXPORT igraph_error_t igraph_community_eb_get_merges(const igraph_t *graph,
                                                 igraph_bool_t directed,
                                                 const igraph_vector_int_t *edges,
                                                 const igraph_vector_t *weights,
                                                 igraph_matrix_int_t *merges,
                                                 igraph_vector_int_t *bridges,
                                                 igraph_vector_t *modularity,
                                                 igraph_vector_int_t *membership);

IGRAPH_EXPORT igraph_error_t igraph_community_fastgreedy(const igraph_t *graph,
                                              const igraph_vector_t *weights,
                                              igraph_matrix_int_t *merges,
                                              igraph_vector_t *modularity,
                                              igraph_vector_int_t *membership);

IGRAPH_EXPORT igraph_error_t igraph_community_to_membership(const igraph_matrix_int_t *merges,
                                                 igraph_int_t nodes,
                                                 igraph_int_t steps,
                                                 igraph_vector_int_t *membership,
                                                 igraph_vector_int_t *csize);
IGRAPH_EXPORT igraph_error_t igraph_le_community_to_membership(const igraph_matrix_int_t *merges,
                                                    igraph_int_t steps,
                                                    igraph_vector_int_t *membership,
                                                    igraph_vector_int_t *csize);

IGRAPH_EXPERIMENTAL IGRAPH_EXPORT igraph_error_t igraph_community_voronoi(
    const igraph_t *graph,
    igraph_vector_int_t *membership, igraph_vector_int_t *generators, igraph_real_t *modularity,
    const igraph_vector_t *lengths, const igraph_vector_t *weights,
    igraph_neimode_t mode, igraph_real_t r);

IGRAPH_EXPORT igraph_error_t igraph_modularity(const igraph_t *graph,
                                    const igraph_vector_int_t *membership,
                                    const igraph_vector_t *weights,
                                    igraph_real_t resolution,
                                    igraph_bool_t directed,
                                    igraph_real_t *modularity);

IGRAPH_EXPORT igraph_error_t igraph_modularity_matrix(const igraph_t *graph,
                                           const igraph_vector_t *weights,
                                           igraph_real_t resolution,
                                           igraph_matrix_t *modmat,
                                           igraph_bool_t directed);

IGRAPH_EXPORT igraph_error_t igraph_reindex_membership(igraph_vector_int_t *membership,
                                            igraph_vector_int_t *new_to_old,
                                            igraph_int_t *nb_clusters);

typedef enum { IGRAPH_LEVC_HIST_SPLIT = 1,
               IGRAPH_LEVC_HIST_FAILED,
               IGRAPH_LEVC_HIST_START_FULL,
               IGRAPH_LEVC_HIST_START_GIVEN
             } igraph_leading_eigenvector_community_history_t;

/**
 * \typedef igraph_community_leading_eigenvector_callback_t
 * Callback for the leading eigenvector community finding method.
 *
 * The leading eigenvector community finding implementation in igraph
 * is able to call a callback function, after each eigenvalue
 * calculation. This callback function must be of \c
 * igraph_community_leading_eigenvector_callback_t type.
 * The following arguments are passed to the callback:
 * \param membership The actual membership vector, before recording
 *    the potential change implied by the newly found eigenvalue.
 * \param comm The id of the community that the algorithm tried to
 *    split in the last iteration. The community IDs are indexed from
 *    zero here!
 * \param eigenvalue The eigenvalue the algorithm has just found.
 * \param eigenvector The eigenvector corresponding to the eigenvalue
 *    the algorithm just found.
 * \param arpack_multiplier A function that was passed to \ref
 *    igraph_arpack_rssolve() to solve the last eigenproblem.
 * \param arpack_extra The extra argument that was passed to the
 *    ARPACK solver.
 * \param extra Extra argument that as passed to \ref
 *    igraph_community_leading_eigenvector().
 *
 * \sa \ref igraph_community_leading_eigenvector(), \ref
 * igraph_arpack_function_t, \ref igraph_arpack_rssolve().
 */

typedef igraph_error_t igraph_community_leading_eigenvector_callback_t(
    const igraph_vector_int_t *membership,
    igraph_int_t comm,
    igraph_real_t eigenvalue,
    const igraph_vector_t *eigenvector,
    igraph_arpack_function_t *arpack_multiplier,
    void *arpack_extra,
    void *extra);

IGRAPH_EXPORT igraph_error_t igraph_community_leading_eigenvector(const igraph_t *graph,
                                                       const igraph_vector_t *weights,
                                                       igraph_matrix_int_t *merges,
                                                       igraph_vector_int_t *membership,
                                                       igraph_int_t steps,
                                                       igraph_arpack_options_t *options,
                                                       igraph_real_t *modularity,
                                                       igraph_bool_t start,
                                                       igraph_vector_t *eigenvalues,
                                                       igraph_vector_list_t *eigenvectors,
                                                       igraph_vector_int_t *history,
                                                       igraph_community_leading_eigenvector_callback_t *callback,
                                                       void *callback_extra);

IGRAPH_EXPORT igraph_error_t igraph_community_fluid_communities(const igraph_t *graph,
                                                     igraph_int_t no_of_communities,
                                                     igraph_vector_int_t *membership);


IGRAPH_EXPORT igraph_error_t igraph_community_label_propagation(const igraph_t *graph,
                                                     igraph_vector_int_t *membership,
                                                     igraph_neimode_t mode,
                                                     const igraph_vector_t *weights,
                                                     const igraph_vector_int_t *initial,
                                                     const igraph_vector_bool_t *fixed,
                                                     igraph_lpa_variant_t variant);

IGRAPH_EXPORT igraph_error_t igraph_community_multilevel(const igraph_t *graph,
                                              const igraph_vector_t *weights,
                                              igraph_real_t resolution,
                                              igraph_vector_int_t *membership,
                                              igraph_matrix_int_t *memberships,
                                              igraph_vector_t *modularity);

typedef enum {
    IGRAPH_LEIDEN_OBJECTIVE_MODULARITY = 0,
    IGRAPH_LEIDEN_OBJECTIVE_CPM,
    IGRAPH_LEIDEN_OBJECTIVE_ER
} igraph_leiden_objective_t;

IGRAPH_EXPORT igraph_error_t igraph_community_leiden(
        const igraph_t *graph,
        const igraph_vector_t *edge_weights,
        const igraph_vector_t *vertex_out_weights,
        const igraph_vector_t *vertex_in_weights,
        igraph_real_t resolution,
        igraph_real_t beta,
        igraph_bool_t start,
        igraph_int_t n_iterations,
        igraph_vector_int_t *membership,
        igraph_int_t *nb_clusters, igraph_real_t *quality);

IGRAPH_EXPORT igraph_error_t igraph_community_leiden_simple(
        const igraph_t *graph,
        const igraph_vector_t *weights,
        igraph_leiden_objective_t objective,
        igraph_real_t resolution,
        igraph_real_t beta,
        igraph_bool_t start,
        igraph_int_t n_iterations,
        igraph_vector_int_t *membership,
        igraph_int_t *nb_clusters,
        igraph_real_t *quality);

/* -------------------------------------------------- */
/* Community Structure Comparison                     */
/* -------------------------------------------------- */

IGRAPH_EXPORT igraph_error_t igraph_compare_communities(const igraph_vector_int_t *comm1,
                                             const igraph_vector_int_t *comm2,
                                             igraph_real_t* result,
                                             igraph_community_comparison_t method);
IGRAPH_EXPORT igraph_error_t igraph_split_join_distance(const igraph_vector_int_t *comm1,
                                             const igraph_vector_int_t *comm2,
                                             igraph_int_t* distance12,
                                             igraph_int_t* distance21);

IGRAPH_END_C_DECLS

#endif
