/*
   igraph library.
   Copyright (C) 2008-2020  The igraph development team
   334 Harvard street, Cambridge, MA 02139 USA

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc.,  51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301 USA

*/

#ifndef IGRAPH_ISOCLASSES_H
#define IGRAPH_ISOCLASSES_H

#include "igraph_decls.h"

IGRAPH_BEGIN_C_DECLS

extern const unsigned int igraph_i_isoclass2_3[];
extern const unsigned int igraph_i_isoclass2_4[];
extern const unsigned int igraph_i_isoclass2_3u[];
extern const unsigned int igraph_i_isoclass2_4u[];
extern const unsigned int igraph_i_isoclass2_5u[];
extern const unsigned int igraph_i_isoclass2_6u[];
extern const unsigned int igraph_i_isoclass_3_idx[];
extern const unsigned int igraph_i_isoclass_4_idx[];
extern const unsigned int igraph_i_isoclass_3u_idx[];
extern const unsigned int igraph_i_isoclass_4u_idx[];
extern const unsigned int igraph_i_isoclass_5u_idx[];
extern const unsigned int igraph_i_isoclass_6u_idx[];

IGRAPH_END_C_DECLS

#endif
